#include "gtk/gtk.h"
#include <iostream>
#include <string>

std::string get_properties(GType gtype)
{
  std::string strResult;
  std::string strObjectName = g_type_name(gtype);

  //Get the list of properties:
  GObjectClass* pGClass = G_OBJECT_CLASS(g_type_class_ref(gtype));

  guint iCount = 0;
  GParamSpec** ppParamSpec = g_object_class_list_properties (pGClass, &iCount);

  //Loop through the list of properties:
  if(ppParamSpec)
  {
    for(guint i = 0; i < iCount; i++)
    {
      GParamSpec* pParamSpec = ppParamSpec[i];
      if(pParamSpec)
      {
        //Name and type:
        std::string strName = g_param_get_name(pParamSpec);
        std::string strTypeName = G_PARAM_SPEC_TYPE_NAME(pParamSpec);

        strResult += "(define-property " + strName + "\n";
        strResult += "  (of-object \"" + strObjectName + "\")\n";
        strResult += "  (prop-type \"" + strTypeName + "\")\n";

        //Flags:
        GParamFlags flags = pParamSpec->flags;
        bool bReadable = (flags & G_PARAM_READABLE) == G_PARAM_READABLE;
        bool bWritable = (flags & G_PARAM_WRITABLE) == G_PARAM_WRITABLE;
        bool bConstructOnly = (flags & G_PARAM_CONSTRUCT_ONLY) == G_PARAM_CONSTRUCT_ONLY;

        //#t and #f aren't documented, but I guess that it's correct based on the example in the .defs spec.
        const std::string strTrue = "#t";
        const std::string strFalse = "#f";

        strResult += "  (readable " + (bReadable ? strTrue : strFalse) + ")\n";
        strResult += "  (writable " + (bWritable ? strTrue : strFalse) + ")\n";
        strResult += "  (construct-only " + (bConstructOnly ? strTrue : strFalse) + ")\n";

        strResult += ")\n\n"; //close (define-property		
      }
    }
  }

  //TODO: free the ppParamSpec? How?
	g_type_class_unref(pGClass); //to match the g_type_class_ref() above.	

  return strResult;
}

std::string get_type_name(GType gtype) //Adds a * if necessary.
{
  std::string strTypeName = g_type_name(gtype);

  if( g_type_is_a(gtype, G_TYPE_OBJECT) || g_type_is_a(gtype, G_TYPE_BOXED) )
    strTypeName += "*";  //Add * to show that it's a pointer.
  else if( g_type_is_a(gtype, G_TYPE_STRING) )
    strTypeName = "gchar*"; //g_type_name() returns "gchararry".

  return strTypeName;
}

std::string get_type_name_parameter(GType gtype) //Adds a * if necessary.
{
  std::string strTypeName = get_type_name(gtype);

  //All signal parameters that are registered as GTK_TYPE_STRING are actually const gchar*.
  if(strTypeName == "gchar*")
    strTypeName = "const-gchar*";

  return strTypeName;
}

std::string get_signals(GType gtype)
{
  std::string strResult;
  std::string strObjectName = g_type_name(gtype);

  //Get the list of signals:
  gpointer gclass_ref = g_type_class_ref(gtype); //Ensures that class_init() is called.

  guint iCount = 0;
  guint* pIDs = g_signal_list_ids (gtype, &iCount);

  //Loop through the list of signals:
  if(pIDs)
  {
    for(guint i = 0; i < iCount; i++)
    {
      guint signal_id = pIDs[i];

      //Name:
      std::string strName = g_signal_name(signal_id);
      strResult += "(define-signal " + strName + "\n";
      strResult += "  (of-object \"" + strObjectName + "\")\n";



      //Other information about the signal:
      GSignalQuery signalQuery = {0};
      g_signal_query(signal_id, &signalQuery);

      //Return type:
      std::string strReturnTypeName = get_type_name_parameter( signalQuery.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE ); //The type is mangled with a flag. Hacky.
      bool bReturnTypeHasStaticScope = (signalQuery.return_type & G_SIGNAL_TYPE_STATIC_SCOPE) == G_SIGNAL_TYPE_STATIC_SCOPE;
      strResult += "  (return-type \"" + strReturnTypeName + "\")\n";


      //When:
      {
        bool bWhenFirst = (signalQuery.signal_flags & G_SIGNAL_RUN_FIRST) == G_SIGNAL_RUN_FIRST;
        bool bWhenLast = (signalQuery.signal_flags & G_SIGNAL_RUN_LAST) == G_SIGNAL_RUN_LAST;

        std::string strWhen;

        if(bWhenFirst && bWhenLast)
          strWhen = "both";
        else if(bWhenFirst)
          strWhen = "first";
        else if(bWhenLast)
          strWhen = "last";

        strResult += "  (when \"" + strWhen + "\")\n";
      }


      //Loop through the list of parameters:
      const GType* pParameters = signalQuery.param_types;
      if(pParameters)
      {
        strResult += "  (parameters\n";

        for(int i = 0; i < signalQuery.n_params; i++)
        {
          GType typeParamMangled = pParameters[i];

          //Parameter name:
          //TODO: How can we get the real parameter name?
          gchar* pchNum = g_strdup_printf("%d", i);
          std::string strParamName = "p" + std::string(pchNum);
          g_free(pchNum);
          pchNum = 0;

          //Just like above, for the return type:
          std::string strTypeName = get_type_name( typeParamMangled & ~G_SIGNAL_TYPE_STATIC_SCOPE ); //The type is mangled with a flag. Hacky.
          bool bReturnTypeHasStaticScope = (typeParamMangled & G_SIGNAL_TYPE_STATIC_SCOPE) == G_SIGNAL_TYPE_STATIC_SCOPE;

          strResult += "    '(\"" + strTypeName + "\" \"" + strParamName + "\")\n";
        }

        strResult += "  )\n"; //close (properties
      }

      strResult += ")\n\n"; //close (define=signal
    }
  }


  g_type_class_unref(gclass_ref); //to match the g_type_class_ref() above.

  //TODO: free the pIDs? How?

  return strResult;
}



std::string get_defs(GType gtype)
{
  std::string strObjectName = g_type_name(gtype);
  std::string strDefs = ";; From " + strObjectName + "\n\n" + get_signals(gtype) + get_properties(gtype);

  return strDefs;
}


int main (int argc, char *argv[])
{
  gtk_init(&argc, &argv);

  std::cout << get_defs( GTK_TYPE_ACCEL_GROUP )
            << get_defs( GTK_TYPE_ACCEL_LABEL )
            << get_defs( GTK_TYPE_ADJUSTMENT )
            << get_defs( GTK_TYPE_ALIGNMENT)
            << get_defs( GTK_TYPE_ARROW )
            << get_defs( GTK_TYPE_ASPECT_FRAME )
            << get_defs( GTK_TYPE_BIN )
            << get_defs( GTK_TYPE_BOX )
            << get_defs( GTK_TYPE_BUTTON )
            << get_defs( GTK_TYPE_BUTTON_BOX )
            << get_defs( GTK_TYPE_CALENDAR )
            << get_defs( GTK_TYPE_CHECK_BUTTON )
            << get_defs( GTK_TYPE_CHECK_MENU_ITEM )
            << get_defs( GTK_TYPE_CLIST )
            << get_defs( GTK_TYPE_COLOR_SELECTION )
            << get_defs( GTK_TYPE_COMBO )
            << get_defs( GTK_TYPE_CONTAINER )
            << get_defs( GTK_TYPE_CTREE )
            << get_defs( GTK_TYPE_CURVE )
            << get_defs( GTK_TYPE_DIALOG )
            << get_defs( GTK_TYPE_DRAWING_AREA )
            << get_defs( GTK_TYPE_ENTRY )
            << get_defs( GTK_TYPE_EVENT_BOX )
            << get_defs( GTK_TYPE_FILE_SELECTION )
            << get_defs( GTK_TYPE_FIXED )
            << get_defs( GTK_TYPE_FONT_SELECTION )
            << get_defs( GTK_TYPE_FONT_SELECTION_DIALOG )
            << get_defs( GTK_TYPE_FRAME )
            << get_defs( GTK_TYPE_HANDLE_BOX )
            << get_defs( GTK_TYPE_IMAGE )
            << get_defs( GTK_TYPE_INPUT_DIALOG )
            << get_defs( GTK_TYPE_INVISIBLE )
            << get_defs( GTK_TYPE_ITEM )
            << get_defs( GTK_TYPE_LABEL )
            << get_defs( GTK_TYPE_LAYOUT )
            << get_defs( GTK_TYPE_LIST )
            << get_defs( GTK_TYPE_LIST_ITEM )
            << get_defs( GTK_TYPE_MENU )
            << get_defs( GTK_TYPE_MENU_BAR )
            << get_defs( GTK_TYPE_MENU_ITEM )
            << get_defs( GTK_TYPE_MENU_SHELL )
            << get_defs( GTK_TYPE_MISC )
            << get_defs( GTK_TYPE_NOTEBOOK )
            << get_defs( GTK_TYPE_OBJECT )
            << get_defs( GTK_TYPE_OPTION_MENU )
            << get_defs( GTK_TYPE_PANED )
            << get_defs( GTK_TYPE_PIXMAP )
            << get_defs( GTK_TYPE_PLUG )
            << get_defs( GTK_TYPE_PREVIEW )
            << get_defs( GTK_TYPE_PROGRESS )
            << get_defs( GTK_TYPE_PROGRESS_BAR )
            << get_defs( GTK_TYPE_RADIO_BUTTON )
            << get_defs( GTK_TYPE_RADIO_MENU_ITEM )
            << get_defs( GTK_TYPE_RANGE )
            << get_defs( GTK_TYPE_RULER )
            << get_defs( GTK_TYPE_SCALE )
            << get_defs( GTK_TYPE_SCROLLBAR )
            << get_defs( GTK_TYPE_SCROLLED_WINDOW )
            << get_defs( GTK_TYPE_SEPARATOR )
            << get_defs( GTK_TYPE_SOCKET )
            << get_defs( GTK_TYPE_SPIN_BUTTON )
            << get_defs( GTK_TYPE_STATUSBAR )
            << get_defs( GTK_TYPE_STYLE )
            << get_defs( GTK_TYPE_TABLE )
            << get_defs( GTK_TYPE_TEAROFF_MENU_ITEM )
            << get_defs( GTK_TYPE_TIPS_QUERY )
            << get_defs( GTK_TYPE_TOGGLE_BUTTON )
            << get_defs( GTK_TYPE_TOOLBAR )
            << get_defs( GTK_TYPE_TOOLTIPS )
            << get_defs( GTK_TYPE_VIEWPORT )
            << get_defs( GTK_TYPE_WIDGET )
            << get_defs( GTK_TYPE_WINDOW );


  return 0;
}
